/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include "Primitive.h"
#include "PrimitiveP.h"
#include "PopupMenu.h"
#include "PopupMenuP.h"

#define W	w -> popupMenu
#define P	w -> primitive
#define C	w -> core

#define ACCELERATOR_OFFSET	30
#define MARGINS	(P.shadow_thickness + W.margin)

/*
 *	Les ressources du menu.
 */

static XtResource resources [] =
{
	{ XtNcursor, XtCCursor, XtRCursor, sizeof (Cursor),
	  XtOffset(PopupMenuWidget, primitive.cursor), XtRString, "arrow"
	},
	{ XtNitem, XtCItem, XtRStringArray, sizeof (String *),
	  XtOffset(PopupMenuWidget, popupMenu.item), XtRImmediate, NULL
	},
	{ XtNfont, XtCFont, XtRFontStruct, sizeof (XFontStruct *),
	  XtOffset(PopupMenuWidget, popupMenu.font), XtRString, "Fixed"
	},
	{ XtNmargin, XtCMargin, XtRInt, sizeof (int),
	  XtOffset(PopupMenuWidget, popupMenu.margin), XtRImmediate, (XtPointer) 2
	},
	{ XtNoblong, XtCOblong, XtRBoolean, sizeof (Boolean),
	  XtOffset (PopupMenuWidget, popupMenu.oblong), XtRImmediate, (XtPointer) False
	},
	{ XtNarmColor, XtCArmColor, XtRPixel, sizeof (Pixel),
	  XtOffset (PopupMenuWidget, popupMenu.arm_color), XtRCallProc, (XtPointer) _GetArmColor
	},
	{ XtNacceleratorColor, XtCForeground, XtRPixel, sizeof (Pixel),
	  XtOffset (PopupMenuWidget, popupMenu.accelerator_color), XtRString, "black"
	},
};


static void		Initialize (Widget, Widget, Arg*, Cardinal*);
static void		Destroy (Widget);
static void		Redisplay (Widget, XEvent*, Region);
static Boolean	SetValues (Widget, Widget, Widget, Arg*, Cardinal*);

static void		Drag (Widget, XEvent*, String*, Cardinal*);

static XtActionsRec actionsList[] =
{
	{"drag", Drag },
};

static char defaultTranslations [] = "\
	<BtnDown>:	drag()\n\
	<BtnMotion>:	drag()\n\
	<BtnUp>:	drag()";

PopupMenuClassRec popupMenuClassRec = {
	/* CoreClassPart */
	{
		(WidgetClass) &primitiveClassRec,	/* superclass */
		"PopupMenu",						/* class_name */
		sizeof (PopupMenuRec),				/* widget_size */
		NULL,								/* class_initialize */
		NULL,								/* class_part_init */
		FALSE,								/* class_inited */
		Initialize,							/* initialize */
		NULL,								/* initialize_hook */
		XtInheritRealize,					/* realize */
		actionsList,						/* actions */
		XtNumber (actionsList),				/* num_actions */
		resources,							/* resources */
		XtNumber (resources),				/* num_resources */
		NULLQUARK,							/* xrm_class */
		True,								/* compress_motion */
		TRUE,								/* compress_exposure */
		TRUE,								/* compress_enterlv */
		FALSE,								/* visible_interest */
		Destroy,							/* destroy */
		NULL,								/* resize */
		Redisplay,							/* expose */
		SetValues,							/* set_values */
		NULL,								/* set_values_hook */
		XtInheritSetValuesAlmost,			/* set_values_almost */
		NULL,								/* get_values_hook */
		NULL,								/* accept_focus */
		XtVersion,							/* version */
		NULL,								/* callback_private */
		defaultTranslations,				/* tm_table */
		NULL,								/* query_geometry */
	},
	/* PrimitiveClassPart */
	{
		NULL,								/* Extension */
	},
	/* PopupMenuClassPart */
	{
		NULL,								/* Extension */
	},
};

WidgetClass popupMenuWidgetClass = (WidgetClass)&popupMenuClassRec;

/*
 *	le GC pour le label.
 */

static void get_labelGC (PopupMenuWidget w)
{
	static Pixmap	greyStipple = (Pixmap)0;
	static char		grey_bits [] = {
		0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55
	};
	XGCValues		values;
	unsigned long	valuemask;

	if (!greyStipple)
		greyStipple = XCreateBitmapFromData (
						XtDisplay ((Widget) w),
						RootWindowOfScreen (XtScreen ((Widget) w)),
						grey_bits, 8, 8);

	valuemask = GCForeground | GCBackground | GCFont;
	values.foreground = P.foreground;
	values.background = C.background_pixel;
	values.font = W.font -> fid;
	W.label_GC = XtGetGC ((Widget) w, valuemask, &values);

	values.fill_style = FillStippled;
	values.stipple = greyStipple;
	valuemask |= GCFillStyle | GCStipple;
	W.label_insensitive_GC = XtGetGC ((Widget) w, valuemask, &values);
}

/*
 *	Trouve la ligne en pixels de l'item 'n' du menu.
 */

static int get_line_from_number (PopupMenuWidget w, int n)
{
	char**	p = W.item;
	int		i;
	int		fh = P.shadow_thickness + W.font_height + 2 * W.margin;
	int		pixline = MARGINS + 1;

	for (i = 1; *p != 0 && i <= n; i++, p++)
		if (**p == '-' || **p == '=')
			pixline += fh/2 - 2 - (P.shadow_thickness % 2);
		else
			pixline += fh;
	return pixline;
}

/*
 *	Trouve l'item de la ligne en pixels 'line' du menu.
 */

static int get_number_from_line (PopupMenuWidget w, int line)
{
	char**	p = W.item;
	int		i;
	int		fh = P.shadow_thickness + W.font_height + 2 * W.margin;
	int		pixline = MARGINS;

	for (i = 0; *p != 0 && pixline < line; i++, p++)
		if (**p == '-' || **p == '=')
			pixline += fh/2 - 2;
		else
			pixline += fh;
	return i-1;
}

/*
 *	Calcule la taille du popupMenu.
 */

static void calc_size (PopupMenuWidget w)
{
	XFontStruct*	fp = W.font;
	char**			p;
	char*			cp;
	int				num_item = 0;
	int				acc_width = 0;
	int				label_width = 0;
	int				aw, lw;
	int				fh = P.shadow_thickness + W.font_height + 2 * W.margin;
	int				ind = 0;
	int				h = MARGINS;

	if (W.item)
		for (p = W.item; *p != 0; p++, num_item++) {
			switch (**p) {
				case '-':
				case '=':
					h += fh/2 - 3;
					continue;

				case 1:		/* togglebutton diamond off */
				case 2:		/* togglebutton diamond on */
				case 3:		/* togglebutton square off */
				case 4:		/* togglebutton square on */
				case 5:		/* insensitive togglebutton */
					ind = W.font_height;
				case 6: 	/* insensitive button */
				case 7: 	/* insensitive button not stippled (title) */

				default:
					if ((cp = index (*p, '\t')) != 0) {
						lw = ind + XTextWidth (fp, *p, cp - *p);
						aw = XTextWidth (fp, cp+1, strlen (cp+1));
					} else {
						lw = ind + XTextWidth (fp, *p, strlen (*p));
						aw = 0;
					}
					h += fh;
			}
			if (label_width < lw)	label_width = lw;
			if (acc_width < aw)		acc_width = aw;
		}

	if (acc_width)
		acc_width += ACCELERATOR_OFFSET;

	C.width = 4 * MARGINS + label_width + acc_width;
	C.height = 2 * MARGINS + h;
	W.acc_pos = label_width + ACCELERATOR_OFFSET;
}

/*
 *	La fonction de reaffichage du menu.
 */

static void item_redraw (PopupMenuWidget w, int n)
{
	char**		p;
	char*		cp;
	char*		pp;
	int			st = P.shadow_thickness;
	int			height = W.font_height + 2 * W.margin;
	int			fh = height + st;
	int			line = st + MARGINS + 1 + W.font -> ascent + (height - W.font_height) / 2;
	int			ind = 0;
	int			i;
	Window		win = XtWindow ((Widget) w);
	Display*	dpy = XtDisplay ((Widget) w);
	GC			gc;

	for (i = 1, p = W.item; *p != 0; i++, p++) {
		switch (**p) {
		case 1: case 2: case 3: case 4:
		case 5:
			ind = W.font_height;
			break;
		case '-':
		case '=':
			if (i <= n)
				line -= fh/2 + 3;
			break;
		}
		if (i <= n)
			line += fh;
	}
	pp = W.item [n];
	gc = W.label_GC;
	/*
	 *	Aligner  droite le dernier mot (premiere tabulation)
	 *	et aligner  gauche le debut de l'item.
	 */
	switch (*pp) {
		case 4:
			_DrawShadow (dpy, win,
							P.top_GC, P.bot_GC, st,
							(ind * 2) / 3 - 2, (ind * 2) / 3 - 2,
							2 * MARGINS,
							line - W.font -> ascent + (ind / 6));
			pp++;
			break;

		case 2:
			XSetForeground (dpy, W.label_GC,
				W.arm_color != 12345 ? W.arm_color : C.background_pixel);
			_DrawDiamondShadow (dpy, win,
					/* x */		2 * MARGINS,
					/* y */		line - W.font -> ascent + (ind / 6),
					/* size */		(ind * 2) / 3,
					/* gc's */		P.top_GC, P.bot_GC, W.label_GC,
					/* fill */		True);
			XSetForeground (dpy, W.label_GC, P.foreground);
			/* fall thru */
		case 1: case 3: case 7:
			pp++;
			break;

		case 5: case 6:
			gc = W.label_insensitive_GC;
			pp++;
			break;
	}
	cp = index (pp, '\t');
	if (cp == NULL || cp [1] == 0)
		XDrawString (dpy, win, gc, ind + 2 * MARGINS, line, pp, strlen (pp));
	else {
		XDrawString (dpy, win, gc, ind + 2 * MARGINS, line, pp, cp - pp);
		XSetForeground (dpy, gc, W.accelerator_color);
		XDrawString (dpy, win, gc,
			2 * MARGINS + W.acc_pos,
			line, cp+1, strlen (cp+1));
		XSetForeground (dpy, gc, P.foreground);
	}
}

/*
 *	Fonction d'initialisation de chaque instance.
 */

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	PopupMenuWidget	w = (PopupMenuWidget) new;

	W.font_height = W.font -> ascent + W.font -> descent + 1;
	W.selected = -1;
	calc_size (w);
	get_labelGC (w);
}

static void hilight_item (PopupMenuWidget w)
{
	int			st = P.shadow_thickness;
	Display*	dpy = XtDisplay ((Widget) w);
	Window		win = XtWindow ((Widget) w);

	if (W.selected == -1 || st == 0)
		return;

	if (!W.oblong) {
		_DrawShadow (dpy, win, P.top_GC, P.bot_GC, st,
						C.width - 2 * MARGINS,
						W.font_height + 2 * MARGINS,
						MARGINS,
						get_line_from_number (w, W.selected));
		return;
	}
	if (W.arm_color != 12345) {
		XSetForeground (dpy, W.label_GC, W.arm_color);
		_FillOvalShadow (dpy, win, W.label_GC,
							C.width - 2 * MARGINS,
							W.font_height + MARGINS,
							MARGINS,
							get_line_from_number (w, W.selected));
		XSetForeground (dpy, W.label_GC, P.foreground);
	}
	_DrawOvalShadow (dpy, win, P.bot_GC, P.top_GC, 1,
						C.width - 2 * MARGINS,
						W.font_height + MARGINS,
						MARGINS,
						get_line_from_number (w, W.selected));
	item_redraw (w, W.selected);
}

/*
 *	Eteint une ligne du popupMenu.
 */

static void clear_item (PopupMenuWidget w)
{
	int	st = P.shadow_thickness;

	if (W.selected == -1 || st == 0)
		return;

	if (!W.oblong) {
		_EraseShadow (XtDisplay ((Widget) w), XtWindow ((Widget) w), st,
						C.width - 2 * MARGINS,
						W.font_height + 2 * MARGINS,
						MARGINS,
						get_line_from_number (w, W.selected));
	} else {
		XClearArea (XtDisplay ((Widget) w), XtWindow ((Widget) w),
						MARGINS,
						get_line_from_number (w, W.selected),
						C.width - MARGINS - W.margin,
						W.font_height + MARGINS + 1,
						False);
		item_redraw (w, W.selected);
	}
}

/*
 *	La fonction de reaffichage.
 */

static void Redisplay (Widget widget, XEvent* event, Region region)
{
	PopupMenuWidget	w = (PopupMenuWidget) widget;
	Display*		dpy = XtDisplay (widget);
	Window			win = XtWindow (widget);
	String*			p;
	String			cp;
	String			pp;
	int				st = P.shadow_thickness;
	int				pixline, line, fh;
	int				height = W.font_height + 2 * W.margin;
	int				ind;
	int				len;
	GC				gc;

	XSetRegion (dpy, P.top_GC, region);
	XSetRegion (dpy, P.bot_GC, region);
	XSetRegion (dpy, W.label_GC, region);
	XSetRegion (dpy, W.label_insensitive_GC, region);

	_DrawShadow (dpy, win,
					P.top_GC, P.bot_GC, st,
					C.width, C.height, 0, 0);

	fh = height + st;
	line = st + MARGINS + 1 + W.font -> ascent + (height - W.font_height) / 2;
	pixline = st + MARGINS + 1 + height / 2;

	for (ind = 0, p = W.item; *p != 0; p++)
		if (1 <= **p && **p <= 5) {
			ind = W.font_height;
			break;
		}

	for (p = W.item; (pp = *p) != 0 ; line += fh, pixline += fh, p++) {
		gc = W.label_GC;
		switch (*pp) {
		case '=':
			_DrawShadow (dpy, win, 
							P.bot_GC, P.top_GC, st,
							C.width - 2 * MARGINS, 2,
							MARGINS, pixline - fh/2 + 3);
			_DrawShadow (dpy, win, 
							P.bot_GC, P.top_GC, st,
							C.width - 2 * MARGINS, 2,
							MARGINS, pixline - fh/2 + 5);
			line -= fh/2 + 3;
			pixline -= fh/2 + 3;
			continue;

		case '-':
			_DrawShadow (dpy, win, 
							P.bot_GC, P.top_GC, st,
							C.width - 2 * MARGINS, 2,
							MARGINS, pixline - fh/2 + 4);
			line -= fh/2 + 3;
			pixline -= fh/2 + 3;
			continue;

		case 4:
			_DrawShadow (dpy, win,
							P.top_GC, P.bot_GC, st,
							(ind * 2) / 3 - 2, (ind * 2) / 3 - 2,
							2 * MARGINS,
							line - W.font -> ascent + (ind / 6));
			pp++;
			break;

		case 2:
			XSetForeground (dpy, W.label_GC,
				W.arm_color != 12345 ? W.arm_color : C.background_pixel);
			_DrawDiamondShadow (dpy, win,
					/* x */		2 * MARGINS,
					/* y */		line - W.font -> ascent + (ind / 6),
					/* size */		(ind * 2) / 3,
					/* gc's */		P.top_GC, P.bot_GC, W.label_GC,
					/* fill */		True);
			XSetForeground (dpy, W.label_GC, P.foreground);
			pp ++;
			break;

		case 1: case 3:
			pp ++;
			break;

		case 5: case 6:
			gc = W.label_insensitive_GC;
			pp ++;
			break;

		case 7:		/* titre: centr */
			len = strlen (++pp);
			XSetForeground (dpy, W.label_GC, W.accelerator_color);
			XDrawString (dpy, win, W.label_GC,
							(C.width - XTextWidth (W.font, pp, len))/2,
							line, pp, len);
			XSetForeground (dpy, W.label_GC, P.foreground);
			continue;

		default:
			break;
		}
		/*
		 *	Aligner  droite le dernier mot (premiere tabulation)
		 *	et aligner  gauche le debut de l'item.
		 */
		cp = index (pp, '\t');
		if (cp == NULL || cp [1] == 0)
			XDrawString (dpy, win, gc, ind + 2 * MARGINS, line, pp, strlen (pp));
		else {
			XDrawString (dpy, win, gc, ind + 2 * MARGINS, line, pp, cp - pp);
			XSetForeground (dpy, gc, W.accelerator_color);
			XDrawString (dpy, win, gc,
				2 * MARGINS + W.acc_pos,
				line, cp+1, strlen (cp+1));
			XSetForeground (dpy, gc, P.foreground);
		}
	}
	if (W.selected != -1)
		item_redraw (w, W.selected);
	/*
	 *	Virer la region de clipping.
	 */
	XSetClipMask (dpy, P.top_GC, None);
	XSetClipMask (dpy, P.bot_GC, None);
	XSetClipMask (dpy, W.label_GC, None);
	XSetClipMask (dpy, W.label_insensitive_GC, None);
}

/*
 *	SetValues.
 */

static Boolean SetValues (Widget current, Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	PopupMenuWidget	cu = (PopupMenuWidget) current;
	PopupMenuWidget	w = (PopupMenuWidget) new;
	Boolean			do_redisplay = False;

	if ((cu -> primitive.shadow_thickness != P.shadow_thickness) ||
		(cu -> popupMenu.item != W.item))
	{
		calc_size (w);
		do_redisplay = True;
	}

	if (cu -> popupMenu.font != W.font) {
		XtDestroyGC (W.label_GC);
		XtDestroyGC (W.label_insensitive_GC);
		get_labelGC (w);
		calc_size (w);
		do_redisplay = True;
	}

	return do_redisplay;
}

/*
 *	La procedure de destruction d'une instance.
 */

static void Destroy (Widget widget)
{
	PopupMenuWidget	w = (PopupMenuWidget) widget;

	XtDestroyGC (W.label_GC);
	XtDestroyGC (W.label_insensitive_GC);
}

static void select_item (PopupMenuWidget w, int cur_item)
{
	char*	label;
	char	c;

	if (cur_item == -1)
		return;

	label = W.item [cur_item];
	if (*label == '-' || *label == '=')
		return;

	if (cur_item != W.selected) {
		if (W.selected != -1) {
			c = W.item [W.selected][0];
			if (c <= 4 || c >= 8)
				clear_item (w);	/* effacer l'ancien */
		}

		c = *label;
		if (c <= 4 || c >= 8) {
			W.selected = cur_item;
			hilight_item (w);	/* allumer le nouveau */
		} else
			W.selected = -1;	/* il est insensible, donc inexistant */
	}
}

/*
 *	Traite les vnements.
 */

static int	the_button;

static void Drag (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	PopupMenuWidget	w = (PopupMenuWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;

	switch (e -> type) {
		case ButtonRelease:
			if (e -> button != the_button)
				return;			/* pas le bon! */

			XtPopdown (XtParent (widget));
			XtUngrabPointer (widget, e -> time);
			XSync (XtDisplay (widget), False);	/* pour s'assurer que la souris n'est */
												/* grabouille quand on appelle la */
												/* callback */
			if (W.selected != -1)
				XtCallCallbackList (widget, P.select, (XtPointer) (long) W.selected);
			W.selected = -1;	/* degager la selection */
			return;

		case MotionNotify:
			/* s'assurer que le menu est devant */
			XRaiseWindow (XtDisplay (widget), XtWindow (XtParent (widget)));
			if (MARGINS < e -> y && e -> y < C.height - MARGINS &&
				MARGINS < e -> x && e -> x < C.width - MARGINS)
			{
				select_item (w, get_number_from_line (w, e -> y - P.shadow_thickness));
				return;
			}
			clear_item (w);	/* effacer l'ancien (on est sorti du menu) */
			W.selected = -1;
			return;
	}
}

/*
 *	Fait bondir le menu.
 */

void PopMenu (Widget widget, XtPointer client, XtPointer call)
{
	XButtonEvent*	e = (XButtonEvent*) call;
	Widget			shell = XtParent ((Widget) client);
	PopupMenuWidget	w = (PopupMenuWidget) client;

	XtVaSetValues (shell,
						XtNx, e -> x_root - 10,
						XtNy, e -> y_root - 10,
						NULL);
	the_button = e -> button;	/* le bouton qui a popp le menu */
	XtPopup (shell, XtGrabNone);
	XtGrabPointer ((Widget) client, False,
					ButtonPressMask|ButtonReleaseMask|ButtonMotionMask,
					GrabModeAsync, GrabModeAsync, None, P.cursor,
					e -> time);
}

Widget CreatePopupMenu (Widget parent, char* name, Arg* args, Cardinal n)
{
	return XtCreateManagedWidget (name, (WidgetClass) &popupMenuClassRec,
			XtVaCreatePopupShell (name, overrideShellWidgetClass, parent,
									"allowShellResize", True,
									NULL),
			args, n);
}
