/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/keysymdef.h>

#include "Primitive.h"
#include "PrimitiveP.h"
#include "ScrollbarP.h"
#include "Scrollbar.h"

/*
 *	Procdures d'action.
 */

static void	Arm (Widget, XEvent*, String*, Cardinal*);
static void	Disarm (Widget, XEvent*, String*, Cardinal*);
static void	Drag (Widget, XEvent*, String*, Cardinal*);

static void move_slider (ScrollbarWidget, int, int);

/* Variables locales */

static int		separationX;
static int		separationY;
static int		initialX;
static int		initialY;
static Boolean	direction;

/* Default translation table and action list  */

static char defaultTranslations [] = "\
	<Btn1Down>:		Arm()\n\
	<Btn1Up>:		Disarm()\n\
	<Btn1Motion>:	Drag()";

static XtActionsRec actionsList [] =
{
	{"Arm",		Arm},
	{"Disarm",	Disarm},
	{"Drag",	Drag},
};

/* Resource list for Scrollbar  */
static XtResource resources[] =
{
	{ XtNshadowThickness, XtCShadowThickness, XtRInt, sizeof (int),
	  XtOffset(ScrollbarWidget, primitive.shadow_thickness), XtRImmediate, (XtPointer) 1
	},
	{ XtNsliderMin, XtCSliderMin, XtRInt, sizeof (int),
	  XtOffset (ScrollbarWidget, scrollbar.min), XtRImmediate, (XtPointer) 0
	},
	{ XtNsliderMax, XtCSliderMax, XtRInt, sizeof (int),
	  XtOffset (ScrollbarWidget, scrollbar.max), XtRImmediate, (XtPointer) 100
	},
	{ XtNsliderOrigin, XtCSliderOrigin, XtRInt, sizeof (int),
	  XtOffset (ScrollbarWidget, scrollbar.origin), XtRImmediate, (XtPointer) 0
	},
	{ XtNsliderExtent, XtCSliderExtent, XtRInt, sizeof (int),
	  XtOffset (ScrollbarWidget, scrollbar.extent), XtRImmediate, (XtPointer) 10
	},
	{ XtNslideOrientation, XtCSlideOrientation, XtRSlideOrientation, sizeof (int),
	  XtOffset (ScrollbarWidget, scrollbar.orientation), XtRImmediate, (XtPointer) XwVERTICAL
	},
	{ XtNsliderMoved, XtCCallback, XtRCallback, sizeof (XtPointer),
	  XtOffset (ScrollbarWidget, scrollbar.slider_moved), XtRPointer, (XtPointer) NULL
	},
	{ XtNareaSelected, XtCCallback, XtRCallback, sizeof (XtPointer),
	  XtOffset (ScrollbarWidget, scrollbar.area_selected), XtRPointer, (XtPointer) NULL
	},
	{ XtNsliderColor, XtCArmColor, XtRPixel, sizeof (Pixel),
	  XtOffset (ScrollbarWidget, scrollbar.slider_color),
	  XtRCallProc, (XtPointer) _GetArmColor
	},
};

/* Static routine definitions  */
static void		Initialize (Widget, Widget, Arg*, Cardinal*);
static void		Resize (Widget);
static void		Destroy (Widget);
static void		Redisplay (Widget, XEvent*, Region);
static Boolean	SetValues (Widget, Widget, Widget, Arg*, Cardinal*);

/* The Scrollbar class record definition  */
ScrollbarClassRec scrollbarClassRec =
{
	{
		(WidgetClass) & primitiveClassRec,	/* superclass */
		"Scrollbar",							/* class_name */
		sizeof (ScrollbarRec),				/* widget_size */
		NULL,								/* class_initialize */
		NULL,								/* class_part_initialize */
		FALSE,								/* class_inited */
		Initialize,							/* initialize */
		NULL,								/* initialize_hook */
		XtInheritRealize,					/* realize */
		actionsList,						/* actions */
		XtNumber (actionsList),				/* num_actions */
		resources,							/* resources */
		XtNumber (resources),				/* num_resources */
		NULLQUARK,							/* xrm_class */
		TRUE,								/* compress_motion */
		XtExposeCompressMaximal,			/* compress_exposure */
		TRUE,								/* compress_enterleave */
		FALSE,								/* visible_interest */
		Destroy,							/* destroy */
		Resize,								/* resize */
		Redisplay,							/* expose */
		SetValues,							/* set_values */
		NULL,								/* set_values_hook */
		XtInheritSetValuesAlmost,			/* set_values_almost */
		NULL,								/* get_values_hook */
		NULL,								/* accept_focus */
		XtVersion,							/* version */
		NULL,								/* callback private */
		defaultTranslations,				/* tm_table */
		NULL,								/* query_geometry */
	},
	{
		NULL,								/* Extension */
	},
	{
		NULL,								/* Extension */
	}
};

WidgetClass scrollbarWidgetClass = (WidgetClass) & scrollbarClassRec;


static void GetSliderGC (ScrollbarWidget w)
{
	Widget		widget = (Widget) w;
	XGCValues	values;

	values.foreground = W.slider_color != 12345
							? W.slider_color
							: P.foreground;
	W.slider_GC = XtGetGC (widget, GCForeground, &values);
}

/*
 *	Calculate the slider location and size in pixels so that it can be drawn.
 */

static void CalcSliderRect (ScrollbarWidget w)
{
	double	userSize;
	double	trueSize;
	double	factor;
	double	slideStart;
	double	slideSize;
	int		borderWidth = P.shadow_thickness;

	userSize = W.max - W.min;
	trueSize = ((W.orientation == XpaHORIZONTAL) ? C.width : C.height) - 2.0 * borderWidth;

	factor = trueSize / userSize;
	slideStart = (double) (W.origin - W.min) * factor + borderWidth;
	slideSize = (double) (W.extent) * factor;

	if (W.orientation == XpaHORIZONTAL) {
		W.sliderX = (int) (slideStart + 0.5);
		W.sliderY = borderWidth;
		W.sliderWidth = (int) (slideSize + 0.5);
		W.sliderHeight = C.height - 2 * borderWidth;
	} else {
		W.sliderX = borderWidth;
		W.sliderY = (int) (slideStart + 0.5);
		W.sliderWidth = C.width - 2 * borderWidth;
		W.sliderHeight = (int) (slideSize + 0.5);
	}
	if (W.sliderWidth < MIN_SLIDER_SIZE) {
		W.sliderWidth = MIN_SLIDER_SIZE;
		if (W.orientation == XpaHORIZONTAL && W.sliderX + W.sliderWidth + 1 > C.width)
			W.sliderX = C.width - W.sliderWidth - 1;
	}
	if (W.sliderHeight < MIN_SLIDER_SIZE) {
		W.sliderHeight = MIN_SLIDER_SIZE;
		if (W.orientation == XpaVERTICAL && W.sliderY + W.sliderHeight + 1 > C.height)
			W.sliderY = C.height - W.sliderHeight - 1;
	}
}

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	ScrollbarWidget	w = (ScrollbarWidget) new;
	Pixel			tmp;

	W.armed = False;
	if (W.min >= W.max) {
		W.min = 0;
		W.max = 100;
	}

	if (W.extent < 1 || W.extent > W.max - W.min)
		W.extent = 1;

	if (W.origin < W.min || W.origin > W.max - W.extent)
		W.origin = W.min;

	if (W.orientation != XpaHORIZONTAL && W.orientation != XpaVERTICAL)
		W.orientation = XpaVERTICAL;

	/* Set up a geometry for the widget if it is currently 0.  */
	if (C.width == 0)
		C.width = (W.orientation == XpaHORIZONTAL) ? 100 : 12;
	if (C.height == 0)
		C.height = (W.orientation == XpaHORIZONTAL) ? 12 : 100;
	/* Get the drawing graphics contexts.  */
	tmp = C.background_pixel;
	C.background_pixel = W.slider_color;
	W.slider_color = tmp;
	GetSliderGC (w);
	CalcSliderRect (w);
}

static void DrawSlider (ScrollbarWidget w, int x, int y)
{
	XFillRectangle (XtDisplay ((Widget) w), XtWindow ((Widget) w),
					W.slider_GC,
					x, y, W.sliderWidth, W.sliderHeight);

	if (P.shadow_thickness > 0)
		_DrawShadow (XtDisplay ((Widget) w), XtWindow ((Widget) w),
						W.armed ? P.bot_GC : P.top_GC,
						W.armed ? P.top_GC : P.bot_GC,
						P.shadow_thickness, W.sliderWidth, W.sliderHeight, x, y);
}

static void Redisplay (Widget widget, XEvent* event, Region region)
{
	ScrollbarWidget	w = (ScrollbarWidget) widget;
	Display*		dpy = XtDisplay (widget);

	XSetRegion (dpy, P.top_GC, region);
	XSetRegion (dpy, P.bot_GC, region);
	XSetRegion (dpy, W.slider_GC, region);
	if (P.shadow_thickness > 0)
		_DrawShadow (dpy, XtWindow (widget),
						P.bot_GC, P.top_GC, P.shadow_thickness,
						C.width, C.height, 0, 0);
	DrawSlider (w, W.sliderX, W.sliderY);
	XSetClipMask (dpy, P.top_GC, None);
	XSetClipMask (dpy, P.bot_GC, None);
	XSetClipMask (dpy, W.slider_GC, None);
}

static void Resize (Widget widget)
{
	CalcSliderRect ((ScrollbarWidget) widget);
}

static void Destroy (Widget widget)
{
	ScrollbarWidget	w = (ScrollbarWidget) widget;

	XtDestroyGC (W.slider_GC);
	XtRemoveAllCallbacks (widget, XtNsliderMoved);
	XtRemoveAllCallbacks (widget, XtNareaSelected);
}

static void RedrawSlider (ScrollbarWidget w)
{
	if (XtIsRealized ((Widget) w)) {
		XClearArea (XtDisplay ((Widget) w), XtWindow ((Widget) w),
					W.sliderX, W.sliderY,
					W.sliderWidth, W.sliderHeight,
					False);
		CalcSliderRect (w);
		DrawSlider (w, W.sliderX, W.sliderY);
	}
}

static Boolean SetValues (Widget current, Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	ScrollbarWidget	cu  = (ScrollbarWidget) current;
	ScrollbarWidget	w = (ScrollbarWidget) new;
	Boolean			returnFlag = False;

	if (W.min >= W.max) {
		W.min = cu->scrollbar.min;
		W.max = cu->scrollbar.max;
	}
	if (W.extent < 1 || W.extent > W.max - W.min) {
		if (cu->scrollbar.extent <= W.max - W.min)
			W.extent = cu->scrollbar.extent;
		else
			W.extent = 1;
	}

	if (W.origin < W.min)
		W.origin = W.min;

	if (W.origin > W.max - W.extent)
		W.origin = W.max - W.extent;

	if (W.orientation != XpaHORIZONTAL && W.orientation != XpaVERTICAL)
		W.orientation = cu->scrollbar.orientation;

	/* See if the GC's need to be regenerated and widget redrawn.  */
	if (P.foreground != cu->primitive.foreground) {
		returnFlag = True;
		XtDestroyGC (W.slider_GC);
		GetSliderGC (w);
	}
	if (W.extent != cu->scrollbar.extent || W.min != cu->scrollbar.min ||
		W.max != cu->scrollbar.max || W.orientation != cu->scrollbar.orientation)
	{
		returnFlag = True;
	}
	if (returnFlag == False && W.origin != cu->scrollbar.origin)
		RedrawSlider (w);

	CalcSliderRect (w);
	return returnFlag;
}


static int CalcSliderVal (ScrollbarWidget w, int x, int y)
{
	int		borderWidth = P.shadow_thickness;
	double	userSize;
	double	trueSize;
	double	referencePoint;
	double	temp;

	if (W.orientation == XpaHORIZONTAL) {
		referencePoint = (double) x - separationX;
		trueSize = C.width - 2 * borderWidth - 4;
	} else {
		referencePoint = (double) y - separationY;
		trueSize = C.height - 2 * borderWidth - 4;
	}
	temp = (referencePoint - borderWidth - 1) / trueSize;
	userSize = W.max - W.min;
	temp = temp * userSize + W.min + 0.5;
	if ((int) temp < W.min)
		temp = W.min;
	else if ((int) temp > W.max - W.extent)
		temp = W.max - W.extent;
	return (int) temp;
}

static XtIntervalId	timer = 0;

static void repeat (XtPointer client, XtIntervalId* id)
{
	ScrollbarWidget	w = (ScrollbarWidget) client;
	XtAppContext	app = XtWidgetToApplicationContext ((Widget) client);

	timer = XtAppAddTimeOut (app, 200, repeat, client);
	if (W.area_selected)
		XtCallCallbackList ((Widget) client, W.area_selected, (XtPointer) (long) direction);
	else
		if (direction)
			move_slider (w, W.sliderX + W.sliderWidth, W.sliderY + W.sliderHeight);
		else
			move_slider (w, W.sliderX - W.sliderWidth, W.sliderY - W.sliderHeight);
}

static void Arm (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	ScrollbarWidget	w = (ScrollbarWidget) widget;
	XButtonEvent*	buttonEvent = (XButtonEvent*) event;
	int				selectionPos;
	int				sliderX = W.sliderX;
	int				sliderY = W.sliderY;
	int				sliderWidth = W.sliderWidth;
	int				sliderHeight = W.sliderHeight;

	initialX = sliderX;
	initialY = sliderY;
	if (buttonEvent->x >= sliderX && buttonEvent->x <= sliderX + sliderWidth - 1 &&
		buttonEvent->y >= sliderY && buttonEvent->y <= sliderY + sliderHeight - 1)
	{
		separationX = buttonEvent->x - sliderX;
		separationY = buttonEvent->y - sliderY;
		W.armed = True;
		DrawSlider (w, sliderX, sliderY);
	} else {
		if (timer)
			XtRemoveTimeOut (timer);
		timer = XtAppAddTimeOut (XtWidgetToApplicationContext (widget),
								400, repeat, (XtPointer) widget);
		direction = (W.orientation == XpaHORIZONTAL)
						? buttonEvent -> x > sliderX
						: buttonEvent -> y > sliderY;
		separationX = 0;
		separationY = 0;
		if (W.area_selected)
			XtCallCallbackList (widget, W.area_selected, (XtPointer) (long) direction);
		else
			if (direction)
				move_slider (w, W.sliderX + W.sliderWidth, W.sliderY + W.sliderHeight);
			else
				move_slider (w, W.sliderX - W.sliderWidth, W.sliderY - W.sliderHeight);
	}
}

static void Disarm (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	ScrollbarWidget	w = (ScrollbarWidget) widget;

	if (timer) {
		XtRemoveTimeOut (timer);
		timer = 0;
	}
	W.armed = False;
	DrawSlider (w, W.sliderX, W.sliderY);
}

static void MoveSlider (ScrollbarWidget w, int currentX, int currentY)
{
	int			oldX = W.sliderX;
	int			oldY = W.sliderY;
	int			width = W.sliderWidth;
	int			height = W.sliderHeight;
	int			st = P.shadow_thickness;
	XSegment	seg [2];

	if (W.orientation == XpaHORIZONTAL) {
		W.sliderX = currentX;
		seg[0].y1 = seg[0].y2 = oldY + 2 * st;
		seg[1].y1 = seg[1].y2 = oldY + height - 3 * st;
		if (oldX < currentX) {
			seg[0].x1 = seg[1].x1 = oldX;
			seg[0].x2 = seg[1].x2 = oldX + currentX - oldX - st;
		} else {
			seg[0].x1 = seg[1].x1 = currentX + width;
			seg[0].x2 = seg[1].x2 = seg[0].x1 + oldX - currentX - st;
		}

		DrawSlider (w, currentX, currentY);
		XClearArea (XtDisplay ((Widget) w), XtWindow ((Widget) w),
					seg[0].x1, oldY,
					seg[0].x2 - seg[0].x1 + st, height, False);
	} else {
		W.sliderY = currentY;
		seg[0].x1 = seg[0].x2 = oldX + 2 * st;
		seg[1].x1 = seg[1].x2 = oldX + width - 3 * st;
		if (oldY < currentY) {
			seg[0].y1 = seg[1].y1 = oldY;
			seg[0].y2 = seg[1].y2 = oldY + currentY - oldY - st;
		} else {
			seg[0].y1 = seg[1].y1 = currentY + height;
			seg[0].y2 = seg[1].y2 = seg[0].y1 + oldY - currentY - st;
		}
		DrawSlider (w, currentX, currentY);
		XClearArea (XtDisplay ((Widget) w), XtWindow ((Widget) w),
					oldX, seg[0].y1,
					width, seg[0].y2 - seg[0].y1 + st, False);
	}
}

static void move_slider (ScrollbarWidget w, int newX, int newY)
{
	int				slideVal;
	int				st = P.shadow_thickness;

	if (W.orientation == XpaHORIZONTAL) {
		newY = W.sliderY;
		if (newX < st)
			newX = st;
		if (newX + W.sliderWidth > C.width - st)
			newX = C.width - W.sliderWidth - st;
	} else {
		newX = W.sliderX;
		if (newY < st)
			newY = st;
		if (newY + W.sliderHeight > C.height - st)
			newY = C.height - W.sliderHeight - st;
	}

	if ((W.orientation == XpaHORIZONTAL && newX != initialX) ||
		(W.orientation == XpaVERTICAL && newY != initialY))
	{
		slideVal = CalcSliderVal (w, newX + separationX, newY + separationY);
		if (slideVal == W.origin)
			return;

		W.origin = slideVal;
		MoveSlider (w, newX, newY);
		XtCallCallbackList ((Widget)w, W.slider_moved, (XtPointer) (long) slideVal);
		initialX = newX;
		initialY = newY;
	}
}

static void Drag (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	ScrollbarWidget	w = (ScrollbarWidget) widget;
	XButtonEvent*	be = (XButtonEvent*) event;

	if (W.armed)
		move_slider (w, be -> x - separationX, be -> y - separationY);
}
