/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>

#include <X11/Intrinsic.h>
#include <X11/keysym.h>

#include <Xpa/Primitive.h>
#include <Xpa/Container.h>
#include <Xpa/ScrolledWin.h>
#include <Xpa/Wix.h>

#include "wix.h"

#define	undo_write(d) (void) write (fdUndoFile, (char*)&(d), sizeof (d))
#define	undo_read(d) (void) read (fdUndoFile, (char*)&(d), sizeof (d))

static int	fdUndoFile = -1;	/* Descripteur du fichier d'UNDO/REDO */

/*
 *	Traite comme d'habitude les vnements qui concernent les
 *	autres fentres que celle qui contient la fentre 'data'
 */

static void get_event_for_widget (WixWindow* data, XEvent* e)
{
	Widget			widget;
	XtAppContext	app = XtWidgetToApplicationContext (data -> active);

	for (;;) {
		XtAppNextEvent (app, e);
		widget = XtWindowToWidget (XtDisplay (data -> pane), e -> xany.window);
		while (widget != NULL) {
			if (widget == data -> toplevel)
				return;
			widget = XtParent (widget);
		}
		XtDispatchEvent (e);
	}
}

/*
 *	Dfait une commande.
 *	retourne:
 *				-1 si on a atteint le dbut du buffer d'UNDO/REDO
 *				l'offset dans le fichier d'undo o est la commande prcdente.
 */

static long undo_one_command (WixWindow* data, long prev)
{
	char	local [2048];
	char*	buffer;
	int		position;
	int		length;
	char	reason;
	long	next;
	long	here = prev;

	/* sauter la commande */
	if (prev >= 0) {
		(void) lseek (fdUndoFile, prev, 0);
		undo_read (prev);
		undo_read (next);
		undo_read (reason);
		/* si ce n'est pas une commande, ne pas la sauter! */
		/* c'est le cas du premier undo lanc  partir du menu */
		if (reason != COMMAND_MARK)
			prev = here;
	}
	/* commencer le dfaisage */
	while (prev >= 0) {
		(void) lseek (fdUndoFile, prev, 0);
		here = prev;
		undo_read (prev);
		undo_read (next);
		undo_read (reason);
		if (reason == COMMAND_MARK) {
			sprintf (data -> diag_file, "Commands to undo: %d", -- data -> undo_number);
			SwSetTitle (XtParent (data -> active), data -> diag_file);
			return here;
		}

		undo_read (position);
		undo_read (length);
		/* dfait une insertion de texte */
		if (reason == INSERT_STRING) {
			(void) lseek (fdUndoFile, (long) length, 1);	/* sauter la chane insre */
			WixReplace (data -> active, position, position + length, NULL, 0);
			continue;
		}
		/* remet  jour le flag 'buffer modifi' */
		if (reason == UNMODIFIED_MARK) {
			modifCallback (data -> active, NULL, NULL);
			continue;
		}
		/* dfait une destruction de texte */
		buffer = length > sizeof (local) ? XtMalloc (length) : local;
		if (length != read (fdUndoFile, buffer, length))
			return here;		/* bizarre ??!? */

		WixReplace (data -> active, position, position, buffer, length);
		if (length > sizeof (local))
			XtFree (buffer);
	}
	XBell (XtDisplay (data -> active), 50);
	modifCallback (data -> active, NULL, NULL);	/* dbut du fichier d'undo/redo => buffer vierge */
	return here;
}

/*
 *	Refait une commande.
 *	retourne:
 *				-1 si on a atteint la fin du buffer d'UNDO/REDO
 *				l'offset dans le fichier d'undo o est la commande suivante.
 */

static long redo_one_command (WixWindow* data, long next)
{
	char	local [2048];
	char*	buffer;
	int		position;
	int		length;
	char	reason;
	long	prev;
	long	here;

	/* sauter la commande */
	(void) lseek (fdUndoFile, next, 0);
	here = next;
	undo_read (prev);
	undo_read (next);
	/* commencer le refaisage */
	while (next >= 0) {
		(void) lseek (fdUndoFile, next, 0);
		here = next;
		undo_read (prev);
		undo_read (next);		/* avancer dans l'historique */
		undo_read (reason);
		if (reason == COMMAND_MARK) {
			sprintf (data -> diag_file, "Commands to undo: %d", ++ data -> undo_number);
			SwSetTitle (XtParent (data -> active), data -> diag_file);
			return here;
		}
		undo_read (position);
		undo_read (length);
		/* refait une destruction de texte */
		if (reason == DELETE_STRING) {
			(void) lseek (fdUndoFile, (long) length, 1);	/* sauter la chane dtruite */
			WixReplace (data -> active, position, position + length, NULL, 0);
			continue;
		}
		/* remet  jour le flag 'buffer modifi' */
		if (reason == UNMODIFIED_MARK) {
			modifCallback (data -> active, NULL, NULL);
			continue;
		}
		/* dfait une insertion de texte */
		buffer = length > sizeof (local) ? XtMalloc (length) : local;
		if (length != read (fdUndoFile, buffer, length))
			break;		/* bizarre ??!? */

		WixReplace (data -> active, position, position, buffer, length);
		if (length > sizeof (local))
			XtFree (buffer);
	}
	XBell (XtDisplay (data -> active), 50);
	return here;
}

/*
 *	L'action d'UNDO/REDO.
 */

void UndoAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	long		here;
	char		tc [32];

	XtVaGetValues (w, XtNuserData, &data, NULL);
	if (data -> undo_ptr == -1 || fdUndoFile == -1) {
		Message (data -> active, "Nothing to undo.");
		return;
	}
	/* Retirer momentanment la scrollbar du bas */
	if (!data -> stripped)
		XtUnmanageChild (SwHSB (XtParent (w)));
	Message (w, "^G:  exit    Up arrow:  Undo    Down arrow:  Redo");
	sprintf (tc, "Number of commands: %d", data -> undo_number);
	SwSetTitle (XtParent (w), tc);
	/* on y va */
	data -> undo_mode = True;
	here = data -> undo_ptr;
	for (;;) {
		char	c;
		KeySym	keysym;
		XEvent	ev;

		/* rcuprer les vnements nous concernant */
		get_event_for_widget (data, &ev);
		switch (ev.type) {
			case Expose:
			case GraphicsExpose:
				XtDispatchEvent (&ev);
				continue;

			case KeyPress:
				if (XLookupString (&ev.xkey, &c, 1, &keysym, 0) == 1 && c == ('g'&037)) {
					data -> undo_ptr = here;
					(void) lseek (fdUndoFile, (long) (sizeof (long) + here), 0);
					here = -1L;
					undo_write (here);
					break;
				}
				if (keysym == XK_Up) {
					here = undo_one_command (data, here);
					WixRedraw (w);
					continue;
				}
				if (keysym == XK_Down) {
					here = redo_one_command (data, here);
					WixRedraw (w);
					continue;
				}
				continue;

			default:
				continue;
		}
		break;
	}
	/* Remettre la scrollbar du bas */
	if (!data -> stripped)
		XtManageChild (SwHSB (XtParent (w)));
	Message (data -> active, "Done.");
	SwSetTitle (XtParent (w), wx_modes [data -> modeNumber].m_name);
	data -> undo_mode = False;
}

/*
 *	La callback qui enregistre les actions de modification rapportes par le widget.
 */

void UndoRedoCallback (Widget w, XtPointer client, XtPointer call)
{
	WixWindow*				data = (WixWindow*) client;
	UndoRedoCallbackStruct*	cb = (UndoRedoCallbackStruct*) call;
	char					reason = cb -> reason;
	long					here;
	/*
	 *	Crer le fichier pour contenir les commandes et les donnes pour UNDO/REDO
	 */
	if (fdUndoFile < 0) {
		char*	tmp = "..tmp..";

		fdUndoFile = creat (tmp, 0600);
		if (fdUndoFile < 0)
			return;				/* pas d'undo/redo, tant pis !! */
		close (fdUndoFile);
		fdUndoFile = open (tmp, 2);	/* rouvrir le fichier en lecture et criture */
		unlink (tmp);			/* ne pas laisser trainer le fichier */
		if (fdUndoFile < 0)
			return;				/* toujours pas d'undo/redo, tant pis !! */
	}

	if (data -> undo_mode ||
		(reason == COMMAND_MARK && data -> prev_reason == COMMAND_MARK))
		return;
	/*
	 *	<prev blk> <next blk> <cmd>[<position> <length> <text>]
	 */
	data -> prev_reason = reason;
	here = lseek (fdUndoFile, 0L, 2);	/* aller au bout du fichier */
	undo_write (data -> undo_ptr);		/* crire le pointeur vers le prcdent bloc */
										/* aller au bloc prcdent */
	if (data -> undo_ptr != -1) {		/* s'il y en a un!! */
		(void) lseek (fdUndoFile, (long) (sizeof (long) + data -> undo_ptr), 0);
		undo_write (here);				/* crire le pointeur sur le bloc suivant */
	}
	data -> undo_ptr = here;			/* nouveau bloc courant */
	(void) lseek (fdUndoFile, 0L, 2);	/* retourner au bout du fichier */
	here = -1L;							/* c'est le dernier pour l'instant */
	undo_write (here);					/* place pour le pointeur sur le bloc suivant */
	undo_write (reason);				/* copier le tag */
	/* c'est tout ce qu'il y a  faire en cas de commande */
	if (reason == COMMAND_MARK) {
		data -> undo_number ++;
		return;
	}
	/* ainsi qu'en cas de dmodification du buffer */
	if (reason == UNMODIFIED_MARK)
		return;

	undo_write (cb -> pos);				/* copier la position */
	undo_write (cb -> length);			/* copier la longueur */
	(void) write (fdUndoFile, cb -> string, cb -> length);	/* copier la chane */
}
