#include <stdio.h>
#include <X11/Intrinsic.h>
#include "jpg_gif.h"

typedef unsigned char byte;
typedef int boolean;

#ifdef WITH_GC
#define XtFree(x) (x) = 0
#endif

#define NEXTBYTE (*ptr++)
#define EXTENSION		0x21
#define IMAGESEP		0x2c
#define TRAILER			0x3b
#define INTERLACEMASK	0x40
#define COLORMAPMASK	0x80

static
int BitOffset = 0,				/* Bit Offset of next code */
	XC = 0,
	YC = 0,						/* Output X and Y coords of current pixel */
	Pass = 0,					/* Used by output routine if interlaced pic */
	OutCount = 0,				/* Decompressor output 'stack count' */
	RWidth,
	RHeight,					/* screen dimensions */
	Width,
	Height,						/* image dimensions */
	LeftOfs,
	TopOfs,						/* image offset */
	BitsPerPixel,				/* Bits per pixel, read from GIF header */
	BytesPerScanline,			/* bytes per scanline in output raster */
	ColorMapSize,				/* number of colors */
	Background,					/* background color */
	CodeSize,					/* Code size, read from GIF header */
	InitCodeSize,				/* Starting code size, used during Clear */
	Code,						/* Value returned by ReadCode */
	MaxCode,					/* limiting value for current code size */
	ClearCode,					/* GIF clear code */
	EOFCode,					/* GIF end-of-information code */
	CurCode,
	OldCode,
	InCode,						/* Decompressor variables */
	FirstFree,					/* First free code, generated per GIF spec */
	FreeCode,					/* Decompressor, next free slot in hash table */
	FinChar,					/* Decompressor variable */
	BitMask,					/* AND mask for data size */
	ReadMask;					/* Code AND mask for current code size */

static boolean Interlace, HasColormap;

static byte *Image;				/* The result array */
/* static byte *RawGIF;			/* The heap array to hold it, raw */
static byte *Raster;			/* The raster data stream, unblocked */

	/* The color map, read from the GIF header */
static byte
	Red[256],
	Green[256],
	Blue[256],
	used[256];
static int numused;

static char *id87 = "GIF87a";
static char *id89 = "GIF89a";

static unsigned long cols[256];
static int numcols;

/* Fetch the next code from the raster data stream.  The codes can be
 * any length from 3 to 12 bits, packed into 8-bit bytes, so we have to
 * maintain our location in the Raster array as a BIT Offset.  We compute
 * the byte Offset into the raster array by dividing this by 8, pick up
 * three bytes, compute the bit Offset into our 24-bit chunk, shift to
 * bring the desired code to the bottom, then mask it off and return it. 
 */

static ReadCode (void)
{
	int RawCode;
	int	ByteOffset;

	ByteOffset = BitOffset / 8;
	RawCode = Raster[ByteOffset] + (0x100 * Raster[ByteOffset + 1]);
	if (CodeSize >= 8)
		RawCode += (0x10000 * Raster[ByteOffset + 2]);
	RawCode >>= (BitOffset % 8);
	BitOffset += CodeSize;
	return (RawCode & ReadMask);
}


static AddToPixel (byte Index)
{
	if (YC < Height)
		*(Image + YC * BytesPerScanline + XC) = Index;

	if (!used[Index]) {
		used[Index] = 1;
		numused++;
	}

/* Update the X-coordinate, and if it overflows, update the Y-coordinate */

	if (++XC == Width) {

/* If a non-interlaced picture, just increment YC to the next scan line. 
 * If it's interlaced, deal with the interlace as described in the GIF
 * spec.  Put the decoded scan line out to the screen if we haven't gone
 * past the bottom of it
 */

		XC = 0;
		if (!Interlace)
			YC++;
		else {
			switch (Pass) {
			case 0:
				YC += 8;
				if (YC >= Height) {
					Pass++;
					YC = 4;
				}
				break;
			case 1:
				YC += 8;
				if (YC >= Height) {
					Pass++;
					YC = 2;
				}
				break;
			case 2:
				YC += 4;
				if (YC >= Height) {
					Pass++;
					YC = 1;
				}
				break;
			case 3:
				YC += 2;
				break;
			default:
				break;
			}
		}
	}
}

/*
 *	Floyd/Steinberg dithering.
 */

/* RANGE forces a to be in the range b..c (inclusive) */
#define RANGE(a,b,c) { if (a < b) a = b;  if (a > c) a = c; }

static void quantize (byte* pic, int w, int h)
{
	byte*	pp;
	int		r1, g1, b1;
	int		*thisline, *nextline, *thisptr, *nextptr, *tmpptr;
	int		i, j, val;
	int		imax = h-1;
	int		jmax = w-1;
	int		r_cmp = (256 / nred) + 1;
	int		g_cmp = (256 / ngreen) + 1;
	int		b_cmp = (256 / nblue) + 1;

	thisline = (int *) XtMalloc (w * 3 * sizeof(int));
	if (!thisline) {
		XtFree ((char*) thisline);
		return;
	}

	nextline = (int *) XtMalloc (w * 3 * sizeof(int));
	if (!nextline) {
		XtFree ((char*) nextline);
		return;
	}
	/* get first line of picture */
	pp = pic;	/* save current line address */
	for (j = w, tmpptr = nextline; --j >= 0; ++pic) {
		*tmpptr++ = (int) Red [*pic];
		*tmpptr++ = (int) Green [*pic];
		*tmpptr++ = (int) Blue [*pic];
	}

	for (i=0; i<h; i++) {
		tmpptr = thisline;  thisline = nextline;  nextline = tmpptr;   /* swap */
		if (i!=imax)   /* get next line */
			for (j=w, tmpptr=nextline; --j >= 0; ++pic) {
				*tmpptr++ = (int) Red [*pic];
				*tmpptr++ = (int) Green [*pic];
				*tmpptr++ = (int) Blue [*pic];
			}

		for (j = 0, thisptr = thisline, nextptr = nextline; j < w; ++j) {
			int r, g, b;

			r1 = *thisptr++;  g1 = *thisptr++;  b1 = *thisptr++;
			RANGE(r1,0,255);  RANGE(g1,0,255);  RANGE(b1,0,255);  
			/* choose actual pixel value */
			val = (((r1) / r_cmp) * ngreen * nblue) +
				  (((g1) / g_cmp) * nblue) +
				  ((b1) / b_cmp);
			/* put it back into array */
			*pp++ = val;
			/* compute color errors */
			r1 -= (int) dither_colormap [0] [val];
			g1 -= (int) dither_colormap [1] [val];
			b1 -= (int) dither_colormap [2] [val];
			/* Add fractions of errors to adjacent pixels */
			if (j!=jmax) {  /* adjust RIGHT pixel */
				thisptr[0] += (r1*7) / 16;
				thisptr[1] += (g1*7) / 16;
				thisptr[2] += (b1*7) / 16;
			}

			if (i!=imax) {	/* do BOTTOM pixel */
				nextptr[0] += (r1*5) / 16;
				nextptr[1] += (g1*5) / 16;
				nextptr[2] += (b1*5) / 16;
				if (j>0) {  /* do BOTTOM LEFT pixel */
					nextptr[-3] += (r1*3) / 16;
					nextptr[-2] += (g1*3) / 16;
					nextptr[-1] += (b1*3) / 16;
				}
				if (j!=jmax) {  /* do BOTTOM RIGHT pixel */
					nextptr[3] += (r1)/16;
					nextptr[4] += (g1)/16;
					nextptr[5] += (b1)/16;
				}
				nextptr += 3;
			}
		}
	}
	XtFree ((char*) thisline);
	XtFree ((char*) nextline);
}

static unsigned char* memGIF (byte* ptr, int filesize, int* pwidth, int* pheight)
{
	Display*	dpy = DisplayOfScreen (TheScreen);
	Colormap	cmap = DefaultColormapOfScreen (TheScreen);
	Visual*		visual = DefaultVisualOfScreen (TheScreen);
	GC			gc = DefaultGCOfScreen (TheScreen);
	Window		root = RootWindowOfScreen (TheScreen);
	Boolean		gif89;
	int			aspect;
	float		normaspect;

	/* The hash table used by the decompressor */
	int Prefix[4096];
	int Suffix[4096];

	/* An output array used by the decompressor */
	int OutCode[1025];
	byte ch, ch1;
	byte *ptr1;
	int i, block;
	FILE *fp;

	BitOffset = 0;
	XC = 0;
	YC = 0;
	Pass = 0;
	OutCount = 0;
	RWidth = 0;
	RHeight = 0;
	Width = 0;
	Height = 0;
	LeftOfs = 0;
	TopOfs = 0;
	BitsPerPixel = 0;
	BytesPerScanline = 0;
	ColorMapSize = 0;
	Background = 0;
	CodeSize = 0;
	InitCodeSize = 0;
	Code = 0;
	MaxCode = 0;
	ClearCode = 0;
	EOFCode = 0;
	CurCode = 0;
	OldCode = 0;
	InCode = 0;
	FirstFree = 0;
	FreeCode = 0;
	FinChar = 0;
	BitMask = 0;
	ReadMask = 0;

	if (DefaultDepthOfScreen (TheScreen) == 1)
		return 0;

/*	RawGIF = ptr; */

	if (!(Raster = (byte *) XtMalloc (filesize)))
		return 0;

	if (! strncmp ((char *) ptr, id87, (size_t) 6))
		gif89 = False;
	else
	if (! strncmp((char *) ptr, id89, (size_t) 6))
		gif89 = True;
	else {
		XtFree (Raster);
#ifdef DEBUG
		fprintf (stderr, "id\n");
#endif
		return 0;
	}
	ptr += 6;

/* Get variables from the GIF screen descriptor */

	ch = NEXTBYTE;
	RWidth = ch + 0x100 * NEXTBYTE;		/* screen dimensions... not used. */
	ch = NEXTBYTE;
	RHeight = ch + 0x100 * NEXTBYTE;

	ch = NEXTBYTE;
	HasColormap = ((ch & COLORMAPMASK) ? True : False);

	BitsPerPixel = (ch & 7) + 1;
	numcols = ColorMapSize = 1 << BitsPerPixel;
	BitMask = ColorMapSize - 1;

	Background = NEXTBYTE;		/* background color... not used. */

	aspect = NEXTBYTE;
	if (aspect) {
		if (!gif89) {
			XtFree (Raster);
#ifdef DEBUG
			fprintf (stderr, "aspect\n");
#endif
			return 0;
		}
		normaspect = (float) (aspect + 15) / 64.0;   /* gif89 aspect ratio */
	}

/* Read in global colormap. */

	if (HasColormap) {
		for (i = 0; i < ColorMapSize; i++) {
			Red[i] = NEXTBYTE;
			Green[i] = NEXTBYTE;
			Blue[i] = NEXTBYTE;
			used[i] = 0;
		}
		numused = 0;
	} else {						/* no colormap in GIF file */
		if (!numcols)
			numcols = 256;
		for (i = 0; i < numcols; i++)
			cols[i] = (unsigned long) i;
	}

	for (;;) {
		block = NEXTBYTE;
		if (block == EXTENSION) {	/* parse extension blocks */
			int i, fn, blocksize, aspnum, aspden;
			/* read extension block */
			fn = NEXTBYTE;
			if (fn == 'R') {	/* GIF87 aspect extension */
				int sbsize;

				blocksize = NEXTBYTE;
				if (blocksize == 2) {
					aspnum = NEXTBYTE;
					aspden = NEXTBYTE;
					if (aspden > 0 && aspnum > 0)
						normaspect = (float) aspnum / (float) aspden;
					else {
						normaspect = 1.0;
						aspnum = aspden = 1;
					}
				}
				else {
					for (i = 0; i < blocksize; i++)
						NEXTBYTE;
				}
				while ((sbsize = NEXTBYTE) > 0) {	/* eat any following data subblocks */
					for (i = 0; i < sbsize; i++)
						NEXTBYTE;
				}
			}
			else if (fn == 0xFE) {	/* Comment Extension */
				int ch, j, sbsize, cmtlen;
				byte *ptr1, *cmt, *cmt1, *sp;
	
				cmtlen = 0;
				ptr1 = ptr;	/* remember start of comments */
				/* figure out length of comment */
				do {
					sbsize = NEXTBYTE;
					cmtlen += sbsize;
					for (j = 0; j < sbsize; j++)
						ch = NEXTBYTE;
				} while (sbsize);
#ifdef DEBUG
				if (cmtlen > 0)		/* build into one un-blocked comment */
					fprintf (stderr, "%*.*s\n", cmtlen, cmtlen, ptr1);
#endif
			}
			else if (fn == 0x01) {	/* PlainText Extension */
				int j, sbsize, ch;
				int tgLeft, tgTop, tgWidth, tgHeight, cWidth, cHeight,
					fg, bg;
	
				sbsize = NEXTBYTE;
				tgLeft = NEXTBYTE;
				tgLeft += (NEXTBYTE) << 8;
				tgTop = NEXTBYTE;
				tgTop += (NEXTBYTE) << 8;
				tgWidth = NEXTBYTE;
				tgWidth += (NEXTBYTE) << 8;
				tgHeight = NEXTBYTE;
				tgHeight += (NEXTBYTE) << 8;
				cWidth = NEXTBYTE;
				cHeight = NEXTBYTE;
				fg = NEXTBYTE;
				bg = NEXTBYTE;
				i = 12;
				for (; i < sbsize; i++)
					NEXTBYTE;	/* read rest of first subblock */
	
				/* read (and ignore) data sub-blocks */
				do {
					j = 0;
					sbsize = NEXTBYTE;
					while (j < sbsize) {
						ch = NEXTBYTE;
						j++;
					}
				} while (sbsize);
			}
			else if (fn == 0xF9) {	/* Graphic Control Extension */
				int j, sbsize;
	
				/* read (and ignore) data sub-blocks */
				do {
					j = 0;
					sbsize = NEXTBYTE;
					while (j < sbsize) {
						NEXTBYTE;
						j++;
					}
				} while (sbsize);
			}
			else if (fn == 0xFF) {	/* Application Extension */
				int j, sbsize;
	
				/* read (and ignore) data sub-blocks */
				do {
					j = 0;
					sbsize = NEXTBYTE;
					while (j < sbsize) {
						NEXTBYTE;
						j++;
					}
				} while (sbsize);
			}
			else {		/* unknown extension */
				int j, sbsize;
	
				/* read (and ignore) data sub-blocks */
				do {
					j = 0;
					sbsize = NEXTBYTE;
					while (j < sbsize) {
						NEXTBYTE;
						j++;
					}
				} while (sbsize);
			}
			continue;
		}
		if (block == TRAILER)
			continue;
		if (block == IMAGESEP)
			break;
		/* erreur */
		XtFree (Raster);
#ifdef DEBUG
		fprintf (stderr, "expecting image_sep: 0x%02x\n", block);
#endif
		return 0;
	}
/* Now read in values from the image descriptor */
	ch = NEXTBYTE;
	LeftOfs = ch + 0x100 * NEXTBYTE;
	ch = NEXTBYTE;
	TopOfs = ch + 0x100 * NEXTBYTE;
	ch = NEXTBYTE;
	Width = ch + 0x100 * NEXTBYTE;
	ch = NEXTBYTE;
	Height = ch + 0x100 * NEXTBYTE;
	Interlace = ((NEXTBYTE & INTERLACEMASK) ? True : False);

/* Tester si on ne veut que la gomtrie */

	if (pwidth != 0) {
		*pwidth = Width;
		if (pheight != 0) {
			*pheight = Height;
			XtFree ((char*) Raster);
			return 0;
		}
	}

/* Note that I ignore the possible existence of a local color map.
 * I'm told there aren't many files around that use them, and the spec
 * says it's defined for future use.  This could lead to an error
 * reading some files. 
 */

/* Start reading the raster data. First we get the intial code size
 * and compute decompressor constant values, based on this code size.
 */
	CodeSize = NEXTBYTE;
	ClearCode = (1 << CodeSize);
	EOFCode = ClearCode + 1;
	FreeCode = FirstFree = ClearCode + 2;
/* The GIF spec has it that the code size is the code size used to
 * compute the above values is the code size given in the file, but the
 * code size used in compression/decompression is the code size given in
 * the file plus one. (thus the ++).
 */
	CodeSize++;
	InitCodeSize = CodeSize;
	MaxCode = (1 << CodeSize);
	ReadMask = MaxCode - 1;
/* Read the raster data.  Here we just transpose it from the GIF array
 * to the Raster array, turning it from a series of blocks into one long
 * data stream, which makes life much easier for ReadCode().
 */
	ptr1 = Raster;
	do {
		ch = ch1 = NEXTBYTE;
		while (ch--)
			*ptr1++ = NEXTBYTE;
		if ((Raster - ptr1) > filesize)
			return (Pixmap) 0;
	} while (ch1);
/* Allocate the X Image */
	Image = (byte *) XtMalloc (Width * Height);
	if (!Image) {
		XtFree (Raster);
		return 0;
	}

	BytesPerScanline = Width;

/* Decompress the file, continuing until you see the GIF EOF code.
 * One obvious enhancement is to add checking for corrupt files here.
 */
	Code = ReadCode ();
	while (Code != EOFCode) {
/* Clear code sets everything back to its initial value, then reads the
 * immediately subsequent code as uncompressed data.
 */
		if (Code == ClearCode) {
			CodeSize = InitCodeSize;
			MaxCode = (1 << CodeSize);
			ReadMask = MaxCode - 1;
			FreeCode = FirstFree;
			CurCode = OldCode = Code = ReadCode ();
			FinChar = CurCode & BitMask;
			AddToPixel (FinChar);
		} else {
/* If not a clear code, then must be data: save same as CurCode and InCode */
			CurCode = InCode = Code;
/*
 * If greater or equal to FreeCode, not in the hash table yet;
 * repeat the last character decoded
 */
			if (CurCode >= FreeCode) {
				CurCode = OldCode;
				OutCode[OutCount++] = FinChar;
			}
/* Unless this code is raw data, pursue the chain pointed to by CurCode
 * through the hash table to its end; each code in the chain puts its
 * associated output code on the output queue.
 */
			while (CurCode > BitMask) {
				if (OutCount > 1024) {
					return (Pixmap) 0;
				}
				OutCode[OutCount++] = Suffix[CurCode];
				CurCode = Prefix[CurCode];
			}
/* The last code in the chain is treated as raw data. */
			FinChar = CurCode & BitMask;
			OutCode[OutCount++] = FinChar;
/* Now we put the data out to the Output routine.
 * It's been stacked LIFO, so deal with it that way...
 */
			for (i = OutCount - 1; i >= 0; i--)
				AddToPixel (OutCode[i]);
			OutCount = 0;
/* Build the hash table on-the-fly. No table is stored in the file. */
			Prefix[FreeCode] = OldCode;
			Suffix[FreeCode] = FinChar;
			OldCode = InCode;
/* Point to the next slot in the table.  If we exceed the current
 * MaxCode value, increment the code size unless it's already 12.  If it
 * is, do nothing: the next code decompressed better be CLEAR
 */
			FreeCode++;
			if (FreeCode >= MaxCode) {
				if (CodeSize < 12) {
					CodeSize++;
					MaxCode *= 2;
					ReadMask = (1 << CodeSize) - 1;
				}
			}
		}
		Code = ReadCode ();
	}
	XtFree (Raster);
	if (pwidth)
		*pwidth = Width;
	if (pheight)
		*pheight = Height;

	return Image;
}

static byte* index2rgb (byte* bpic, int width, int height)
{
	byte*	rgbpic = (byte*) XtMalloc (width*height*3);
	byte*	p = rgbpic;
	int		i;

	for (i = width*height; --i >= 0; ++bpic) {
		*p++ = Red [*bpic];
		*p++ = Green [*bpic];
		*p++ = Blue [*bpic];
	}
	return rgbpic;
}

byte* Gif2ps (char* fname, int* pwidth, int* pheight)
{
	FILE*	fp;
	byte*	ptr;
	byte*	rgbpic;
	byte*	bpic;
	int		filesize;

	/* charger le fichier */
	fp = fopen (fname, "r");
	if (!fp)
		return 0;

	/* find the size of the file */
	fseek (fp, 0L, 2);
	filesize = ftell (fp);
	fseek (fp, 0L, 0);

	if (!(ptr = (byte *) XtMalloc (filesize)))
		return 0;

	if (fread (ptr, filesize, 1, fp) != 1) {
		XtFree (ptr);
		return 0;
	}
	fclose (fp);

	bpic = memGIF (ptr, filesize, 0, 0);
	rgbpic = index2rgb (bpic, Width, Height);
	*pwidth = Width;
	*pheight = Height;
	XtFree ((char*) bpic);
	return rgbpic;
}

XImage* Gif2XImage (char* fname, Pixel bg, int* pwidth, int* pheight)
{
	FILE*	fp;
	byte*	ptr;
	byte*	rgbpic;
	byte*	bpic;
	int		filesize;
	XImage*	image;
	Visual*	visual = DefaultVisualOfScreen (TheScreen);
	Boolean	do_quantize = (visual -> class == PseudoColor);
	static Cache*	cache = (Cache*)0;
	Cache*			p;

	if (bg == 0)
		bg = WhitePixelOfScreen (TheScreen);
	/* chercher d'abord dans le cache */
	for (p = cache; p; p = p -> next)
		if (fname [0] == p -> name [0] && !strcmp (fname+1, p -> name+1)) {
			if (pwidth != 0) {
				*pwidth = p -> width;
				*pheight = p -> height;
			}
			return p -> image;
		}
	/* charger le fichier */
	fp = fopen (fname, "r");
	if (!fp) {
		return 0;
	}

	/* find the size of the file */
	fseek (fp, 0L, 2);
	filesize = ftell (fp);
	fseek (fp, 0L, 0);

	if (!(ptr = (byte *) XtMalloc (filesize))) {
		return 0;
	}

	if (fread (ptr, filesize, 1, fp) != 1) {
		XtFree (ptr);
		return 0;
	}
	fclose (fp);

	bpic = memGIF (ptr, filesize, pwidth, pheight);
	/* Juste la gomtrie?? */
	if (pwidth != 0) {
		XtFree (ptr);
		return (XImage*) 0;
	}
	/* Faire une XImage */
	if (do_quantize) {
		initColorCube (bg);
		quantize (bpic, Width, Height);
		image = rgbpic2XImage (bpic, Width, Height, bg);
	} else {
		rgbpic = index2rgb (bpic, Width, Height);
		image = rgbpic2XImage (rgbpic, Width, Height, bg);
		XtFree ((char*) rgbpic);
	}
	XtFree ((char*) bpic);
	XtFree (ptr);
	/* Mettre la nouvelle image dans le cache */
	p = (Cache*) XtMalloc (sizeof (Cache));
	p -> image = image;
	p -> name = fname;
	p -> width = Width;
	p -> height = Height;
	p -> next = cache;
	cache = p;
	/* et voila */
	return p -> image;
}
