/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <Xpa/Primitive.h>
#include <Xpa/Container.h>
#include <Xpa/PushButton.h>
#include <Xpa/PanedWin.h>
#include <Xpa/Wix.h>

#include "wix.h"

#define	isdigit(c) ((c)<='9' && (c)>='0')

static	String	errorTable [] = {
	"Error during macro execution",
	"Beginning of buffer",
	"",
	"End of buffer",
	"No opening parenthesis",
	"No closing parenthesis",
	"Mark Set",
	"Undefined Command",
	"Malformed octal number.",
	"No such font",
	"Expand from the alias table",
	"Expand from the current buffer",
	"Expand from the abbrev file",
	"Expand from the kill stack",
};

/*
 *	Traite les messages d'erreur.
 */

void showErrors (Widget wix, int type, int n)
{
	Boolean	ro;
	char	tc [64];

	switch (type) {
	case T_OK:	Message (wix, NULL);
				return;

	case T_ARGCNT:	sprintf (tc, "Arg Count: %d", n);
					goto show;

	case T_OCTAL:	if (n == 0)
						strcpy (tc, "Octal:");
					else
						sprintf (tc, "Octal: %o", n);
					goto show;

	case T_PREFIX:	if (n == 033)
						strcpy (tc, "ESC-");
					else
						sprintf (tc, "^%c", n + '@');
					goto show;

	case T_ERROR:	Message (wix, errorTable [n]);
					return;

	case T_REMOVED:
			XtVaGetValues (wix, XtNreadOnly, &ro, NULL);
			if (ro || n <= 1) {
				Message (wix, NULL);
				return;
			}
			sprintf (tc, "%d chars removed", n);
			goto show;

	case T_COPIED:	sprintf (tc, "%d chars copied", n);
					goto show;

	case T_WRITTEN:	sprintf (tc, "%d chars written", n);
	show:			Message (wix, tc);
					return;
	}
}

/*
 *	Excute une commande shell.
 */

static void execCommandCont (WixWindow* data)
{
	if (fork ())
		return;

	execl ("/bin/sh", "sh", "-c", data -> fs_answer, 0);
	exit (1);
}

void execCommandAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	data -> what_after = execCommandCont;
	popQueryStringNoWait (data, "Shell command", "", False);
}

/*
 *	Change la commande de compilation du mode courant.
 */

static void setCompileCommandCont (WixWindow* data)
{
	if (wx_modes [data -> modeNumber].m_compile)
		XtFree (wx_modes [data -> modeNumber].m_compile);

	wx_modes [data -> modeNumber].m_compile = XtNewString (data -> fs_answer);
}

void setCompileCommandAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	static char	tc [64];

	XtVaGetValues (w, XtNuserData, &data, NULL);
	data -> what_after = setCompileCommandCont;
	sprintf (tc, "Compilation command for mode\n%s", wx_modes [data -> modeNumber].m_name);
	popQueryStringNoWait (data, tc, wx_modes [data -> modeNumber].m_compile, False);
}

/*
 *	Exploite le fichier d'erreurs.
 */

static void CCerrors (WixWindow*);

static void sigchld () {
	int			status;
	int			pid;
	int			retcode;
	WixWindow*	wp;

	pid = wait (&status);
	for (wp = _head; wp; wp = wp -> next)
		if (wp -> diag_pid == pid) {
			TextCursor (wp);
			XtSetSensitive (wp -> pane, True);
			XSync (XtDisplay (wp -> pane), False);

			retcode = WIFEXITED (status)
				 ? WEXITSTATUS (status)
				 : -1;

			if (retcode != 0) {
				XtAppContext	app = XtWidgetToApplicationContext (wp -> pane);
				char			tc [64];

				CCerrors (wp);
				sprintf (tc, "Process exit code: %d (failure).", retcode);
				Message (wp -> wix, tc);
				while (XtAppPending (app)) {
					XEvent	e;

					XtAppNextEvent (app, &e);
					XtDispatchEvent (&e);
				}
			} else {
				Message (wp -> wix, "Process exit code: 0 (no errors).");
				if (wp -> diag_list) {
					XtUnmanageChild (XtParent (wp -> diag_list));
					ClearDiagBox (wp);
				}
				XSync (XtDisplay (wp -> pane), False);
			}
			return;
		}
}

extern char* get_executable_path (char*);	/* from menu.c */

static void runCC (WixWindow* data, char* p, char* dirname)
{
	char	executable [128];
	char*	pe = executable;
	char*	cp;
	char	file [32];

	/*
	 *	Vrifier que la commande se trouve dans le PATH.
	 */
	for (cp = p;;) {
		switch (*pe ++ = *cp++) {
			case ' ':
			case '\t':
			case 0:
				pe [-1] = 0;
				break;

			default:
				continue;
		}
		break;
	}
	pe = get_executable_path (executable);
	if (pe == NULL) {
		char	tc [512];

		TextCursor (data);
		sprintf (tc, "Your compilation command (%s) has not\nbeen found in the current search path.\nTry the absolute pathname for '%s'",
					executable, executable);
		popAlert (data, tc, False);
		WaitForResponse (data);
		return;
	}

	if ((data -> diag_pid = fork ()) != 0) {
		/*
		 *	Initialiser le fichier de diagnostics.
		 */
		sprintf (data -> diag_file, "/tmp/.wx-diag#%d", data -> diag_pid);
		XtSetSensitive (data -> pane, False);
		return;
	}
	chdir (dirname);
	close (1);
	close (2);
	sprintf (file, "/tmp/.wx-diag#%d", getpid ());
	creat (file, 0666);
	dup (1);
	close (0);
	/*
	 *	Executer la commande de compilation.
	 */
	execl ("/bin/sh", "sh", "-c", p, NULL);
	exit (1);
}

/*
 *	Compile le buffer.
 */

static void compileBufferCont (WixWindow* data, char* cs)
{
	char*			s;
	char			tc [256], temp [256];
	char*			extension;
	char*			basename;
	char*			dirname;
	int				baselen, extlen;
	int				i, dot, slash;

	signal (SIGCHLD, sigchld);
	/*
	 *	Separer le nom de l'extension du directory.
	 */
	slash = dot = -1;
	for (i = 0;; i++) {
		switch (temp [i] = data -> filename [i]) {
			case '/':	slash = i; continue;
			case '.':	dot = i; continue;
			case 0:		break;
			default:	continue;
		}
		break;
	}
	if (temp [0] == '/')	dirname = temp;
	else			dirname = data -> directory;

	if (slash >= 0)	{
		temp [slash] = 0;
		if (dirname [0] == 0)
			dirname = "/";
	}
	basename = temp + ++slash;

	if (dot >= 0) {
		baselen = dot - slash;
		extlen = i - dot -1;
		temp [dot] = 0;
		extension = temp + dot + 1;
	} else {
		baselen = i - slash;
		extlen = 0;
		extension = temp + i;	/* chaine vide */
	}
	/*
	 *	Analyser le format de la commande de compilation.
	 */
	while (*cs == ' ' || *cs == '\t')
		cs ++;		/* sauter les espaces initiaux */
	for (s = tc; *cs; cs++)
		if ((*s++ = *cs) == '%')
			switch (*++cs) {
			case 'b':
			case 'B':
				strcpy (--s, basename);
				s += baselen;
				continue;
			case 'e':
			case 'E':
				strcpy (--s, extension);
				s += extlen;
				continue;
			}
	*s = 0;
	WatchCursor (data);
	XSync (XtDisplay (data -> pane), False);
	runCC (data, tc, dirname);
}

void compileBufferAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	Boolean		m;

	XtVaGetValues (w,
		XtNbufferModified,	&m,
		XtNuserData,		&data,
		NULL);

	if (*num_params == 0 && !wx_modes [data -> modeNumber].m_compile) {
		static char	tc [64];

		sprintf (tc, "No compile command defined\nfor the '%s' mode.",
					wx_modes [data -> modeNumber].m_name);
		popAlert (data, tc, False);
		return;
	}

	if (m)
		saveFileAction (w, NULL, NULL, NULL);
	compileBufferCont (data, *num_params > 0 ? params [0] : wx_modes [data -> modeNumber].m_compile);
}

/*
 *	La commande d'impression.
 */

void printBufferAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	Boolean		m;
	char		tc [128];

	XtVaGetValues (w,
					XtNbufferModified,	&m,
					XtNuserData,		&data,
					NULL);

	if (*num_params == 0 && !wx_modes [data -> modeNumber].m_print) {
		static char	tc [64];

		sprintf (tc, "No print command defined\nfor the '%s' mode.",
					wx_modes [data -> modeNumber].m_name);
		popAlert (data, tc, False);
		return;
	}
	/*
	 *	Demander confirmation.
	 */
	data -> what_after = NULL;
	sprintf (tc, "Print  %s ?", data -> filename);
	popAlert (data, tc, True);
	WaitForResponse (data);
	if (data -> cancelled)
		return;

	if (m)
		saveFileAction (w, NULL, NULL, NULL);
	compileBufferCont (data, *num_params > 0 ? params [0] : wx_modes [data -> modeNumber].m_print);
}

/*
 *	Exploite le fichier d'erreurs.
 */

static void CCerrors (WixWindow* data)
{
	char	buffer [2048];
	FILE*	fp;
	int		pos;
	char*	fmt;

	if (!data -> diag_list)
		CreateDiagBox (data -> pane, data);
	/*
	 *	Effacer les anciens messages.
	 */
	ClearDiagBox (data);
	/*
	 *	Analyser le fichier d'erreurs.
	 */
	if (NULL == (fp = fopen (data -> diag_file, "r")))
		return;

	pos = 0;
	if ((fmt = wx_modes [data -> modeNumber].m_errors) != 0)
	{
		int	line;

		while (NULL != fgets (buffer, sizeof (buffer), fp)) {
			if (0 < sscanf (buffer, fmt, &line)) {
				pos ++;
				WixGotoLine (data -> active, line);
				WixCreateMark (data -> active, (void*) (long) (pos+1024));
				KInsertString (data -> diag_list, (unsigned char*) buffer, strlen (buffer));
			}
		}
	}
	if (pos > 0) {
		XtVaSetValues (data -> diag_list, XtNuserData, pos, NULL);
		PopDiagBox (data);
	} else {
		XtUnmanageChild (XtParent (data -> diag_list));
		ClearDiagBox (data);
	}

	fclose (fp);
	unlink (data -> diag_file);
}

/*
 *	Affiche le kill stack.
 */

static Widget	KillStackWindow = NULL;

static void yank (Widget w, XtPointer client, XtPointer call)
{
	Widget			info = (Widget) client;
	XButtonEvent*	e = (XButtonEvent*) call;
	int				l, len;
	char*			buffer;
	static char		tc [32];

	XtVaGetValues (w, XtNcurrentLineNumber, &l, NULL);

	if (e -> button == Button1) {
		buffer = GetKilled (l+1, (long*) &len);
		XtVaSetValues (info, XtNreadOnly, False, NULL);
		WixClearBuffer (info);
		if (buffer) {
			WixSetSelection (w, e, buffer, len);
			/*
			 *	Afficher le contenu dans le panneau du bas.
			 */
			WixInsert (info, "\t\t");
			WixSetFont (info, 2);
			WixInsert (info, "Current selection\n");
			WixSetFont (info, 0);
			WixInsert (info, "\n");
			WixInsert (info, buffer);
			XtFree (buffer);
			sprintf (tc, "%d chars selected.", len);
			SwSetLabel (XtParent (w), tc);
		} else
			SwSetLabel (XtParent (w), NULL);

		XtVaSetValues (info,
			XtNreadOnly,	True,
			XtNnoCursor,	True,
			XtNbufferModified, False,
			NULL);
		BeginningOfBuffer (info, 1, NULL);
		WixRedraw (info);
	}
}

static void hide (Widget w, XtPointer client, XtPointer call)
{
	XtUnmapWidget ((Widget) client);
}

static void updateKillStack (Widget w, XtPointer client, XEvent* event, Boolean* cont)
{
	long		n;
	char		tmp [1024];
	char*		tc;
	char*		flg;
	int			size, nb;
	static long	currentEndOfKillStack = 0;

	if (currentEndOfKillStack == EndOfKillStack)
		return;

	currentEndOfKillStack = EndOfKillStack;
	XtVaSetValues (KillStackWindow, XtNreadOnly, False, NULL);
	WixClearBuffer (KillStackWindow);

	lseek (FdKillFile, EndOfKillStack, 0);		/* aller au bout */
	for (nb = 1;; nb++) {
		lseek (FdKillFile, (long) -sizeof n, 1);
		if ((sizeof n != read (FdKillFile, (char*)&n, sizeof n)) || (-1 == lseek (FdKillFile, (long) -(n + sizeof n), 1)))
			break;
		sprintf (tmp, "%4d: ", nb);
		tc = tmp + 6;	/* 4 chiffres : espace */
		size = read (FdKillFile, tc, n > 127 ? 127 : n);
		lseek (FdKillFile, (long) -size, 1);
		flg = & tc [size]; *flg = '\n';
		while (*tc++ != '\n');
		if (flg != tc-1) {
			tc [-1] = ' ';
			*tc++ = '.';
			*tc++ = '.';
			*tc++ = '.';
		} else	--tc;
		*tc++ = '\n';
		*tc = '\0';
		KInsertString (KillStackWindow, (unsigned char*) tmp, tc - tmp);
	}
	XtVaSetValues (KillStackWindow,
		XtNreadOnly,	True,
		XtNnoCursor,	True,
		XtNbufferModified, False,
		NULL);
	BeginningOfBuffer (KillStackWindow, 1, NULL);
	WixRedraw (KillStackWindow);
}

void showKillStackAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	Widget	top;
	Widget	pane;
	Arg		arglist [10];
	int		n = 0;

	if (KillStackWindow) {
		XtMapWidget (XtParent (XtParent (XtParent (KillStackWindow))));
		return;
	}
	/*
	 *	Creer le shell.
	 */
	top = XtVaCreatePopupShell ("clipboard", topLevelShellWidgetClass, ghost,
								XtNtitle,		"Wix Clipboard",
								XtNiconName,	"Wix Clipboard",
								NULL);
	/*
	 *	la paned window.
	 */
	pane = XtVaCreateManagedWidget ("pane", panedWindowWidgetClass, top, NULL);
	/*
	 *	Creer le wix liste.
	 */
	n = 0;
	XtSetArg (arglist [n], XtNtranslations, listTrans); n++;
	XtSetArg (arglist [n], XtNnoCursor, True); n++;
	XtSetArg (arglist [n], XtNlineNum, 8); n++;
	XtSetArg (arglist [n], XtNcolumnNumber, 80); n++;
	KillStackWindow = XtCreateScrolledWix (pane, "clipboard", arglist, n);
	XtManageChild (XtParent (KillStackWindow));
	/*
	 *	Creer le wix info.
	 */
	n = 0;
	XtSetArg (arglist [n], XtNtranslations, HelpTrans); n++;
	XtSetArg (arglist [n], XtNnoCursor, True); n++;
	XtSetArg (arglist [n], XtNlineNum, 8); n++;
	XtSetArg (arglist [n], XtNcolumnNumber, 80); n++;
	w = XtCreateScrolledWix (pane, "clipinfo", arglist, n);
	XtManageChild (XtParent (w));
	/*
	 *	Le bouton pour sortir.
	 */
	XtVaSetValues (XtParent (KillStackWindow),
		XtNpixmap, OkPixmap,
		NULL);

	XtAddCallback (XtParent (KillStackWindow), XtNselect, hide, top);
	XtAddCallback (KillStackWindow, XtNactivateCallback, yank, w);
	XtAddEventHandler (KillStackWindow, ExposureMask | EnterWindowMask, False, updateKillStack, NULL);

	XtRealizeWidget (top);
	XtMapWidget (top);
}
